#This sets up the training nets with verb, noun, prep, noun input, and the
#appropriate attachment decision.  It presents the four
#tuples with the binding, and writes the spikes out.
import sys
import numpy as np
import cPickle as pickle

import nealParams as nealParameters
if (nealParameters.simulator=="spinnaker"):
    import pyNN.spiNNaker as sim
elif (nealParameters.simulator=="nest"):
    import pyNN.nest as sim

from nealCoverClass import NealCoverFunctions
from createHierClass import CreateHierClass

#constants
neuronsPerCA = 10
intraCAWeight = 0.005
hierWeight = 0.003
dirName = ""#"inputs/"
timeBetweenPresentations = 200

#global variables
numCAs = -1
preps = list()
numberPreps = -1

def getTrainSpikeTimes(start, step, steps):
    extraTime = 0
    inputSpikeTimes = []
    for numTrainSpikes in range (0,steps): 
        nextTime = start + (numTrainSpikes *step) + extraTime
        inputSpikeTimes = inputSpikeTimes + [nextTime]

    #print inputSpikeTimes
    return inputSpikeTimes

def createInput(startTime,numTrainingInputs,timeBetweenSteps):
    presentTimes = getTrainSpikeTimes(startTime+5,timeBetweenSteps,numTrainingInputs)
    presentArray = {'spike_times': [presentTimes]}
    primaryExciteSpikeGen=sim.Population(1,sim.SpikeSourceArray,presentArray)
    #print presentTimes
    return primaryExciteSpikeGen

def createCAInputs(numCAs):
    spikeSources = []
    timeBetweenEpochs = (numCAs)*timeBetweenPresentations
    startTime = 10
    for CANumber in range (0,numCAs):
        newSource = createInput(startTime,2,timeBetweenEpochs) 
        startTime = startTime + timeBetweenPresentations
        spikeSources = spikeSources + [newSource]

    return spikeSources

def makeSynapsesFromSourceToCA(spikeSource,cells,toStart):
    inputWeight = 0.1
    connector = []
    for toOffset in range (0,neuronsPerCA):
        toNeuron = toOffset +toStart
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, cells, connector,'excitatory')

def makeSynapsesFromSpikeSourcesToCAs(cells):
    popSize = cells.size
    spikeSources = createCAInputs(numCAs)
    for CA in range (0,numCAs):
            makeSynapsesFromSourceToCA(spikeSources[CA],cells,CA*10)

def makeCAs(cells):
    numberNeurons = cells.size
    numberCAs = numberNeurons/neuronsPerCA
    connector = []
    for CA in range (0,numberCAs):
        for fromOffset in range (0,neuronsPerCA):
            for toOffset in range (0,neuronsPerCA):
                if (fromOffset != toOffset):
                    fromNeuron = fromOffset + (CA*neuronsPerCA)
                    toNeuron = toOffset + (CA*neuronsPerCA)
                    connector = connector + [(fromNeuron,toNeuron,intraCAWeight,
                                              nealParameters.DELAY)]

    neal.nealProjection(cells, cells, connector,'excitatory')

def makeHier(fromCA, toCA, cells):
    connector = []
    for fromOffset in range (0,neuronsPerCA):
        for toOffset in range (0,neuronsPerCA):
            fromNeuron = fromOffset + (fromCA*neuronsPerCA)
            toNeuron = toOffset + (toCA*neuronsPerCA)
            connector = connector + [(fromNeuron,toNeuron,hierWeight,
                                              nealParameters.DELAY)]

    neal.nealProjection(cells, cells, connector,'excitatory')

#This uses the createHierClass.  Consider, probably should be
#integrated better.
def makeHiersFromHier(hier):
    for pairNumber in range (0,hier.numberIsAs):
        subCatName = hier.isARelationships[pairNumber][0]
        superCatName = hier.isARelationships[pairNumber][1]
        subCatNumber = hier.getUnitNumber(subCatName)
        superCatNumber = hier.getUnitNumber(superCatName)
        print subCatName,superCatName,subCatNumber,superCatNumber
        makeHier(subCatNumber,superCatNumber,hier.cells)


def makeHiers(cells):
    makeHier(1,0,cells)
    makeHier(2,0,cells)
    makeHier(3,1,cells)
    makeHier(4,1,cells)
    makeHier(5,3,cells)
    makeHier(6,3,cells)

def getStopTrainSpikeTimes(start,steps,stepLength):
    inputSpikeTimes = []
    for numTrainSpikes in range (0,steps): 
        nextTime = start + (numTrainSpikes *stepLength) 
        inputSpikeTimes = inputSpikeTimes + [nextTime]

    #print inputSpikeTimes
    return inputSpikeTimes

def sourceStopsPopulation(spikeSource,cells,popSize):
    if (nealParameters.simulator=="spinnaker"):
        inputWeight = 1.0
    elif (nealParameters.simulator=="nest"):
        inputWeight = -1.0
    connector = []
    for toOffset in range (0,popSize):
        toNeuron = toOffset
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, cells, connector,'inhibitory')

def getPrimeTimes(start,numCAs):
    numTests = numCAs
    spikesPerTest = 7
    timeBetweenSpikes = 7
    timeBetweenTests = timeBetweenPresentations
    inputSpikeTimes = []
    for test in range (0,numTests): 
        time = start + (test*timeBetweenTests)
        for spikeThisTest in range (0,spikesPerTest): 
           nextTime = time + (spikeThisTest*timeBetweenSpikes)
           inputSpikeTimes = inputSpikeTimes + [nextTime]

    #print inputSpikeTimes
    return inputSpikeTimes

def getPrimeSource(start,numCAs):
    primeTimes = getPrimeTimes(start,numCAs)
    primeArray = {'spike_times': [primeTimes]}
    spikeGen=sim.Population(1,sim.SpikeSourceArray,primeArray)
    return spikeGen

#Test the hierarchy
def makeSynapsesFromSpikeSourcesToHiers(cells,hierarchyTestStart):
    popSize = cells.size
    spikeSource = getPrimeSource(hierarchyTestStart+15,numCAs)
    primeWeight = 0.02
    connector = []
    for toNeuron in range (0,popSize):
        connector = connector + [(0,toNeuron,primeWeight,nealParameters.DELAY)]

    neal.nealProjection(spikeSource, cells, connector,'excitatory')


def printPklSpikes(fileName,outFileName):
        fileHandle = open(fileName)
        outFileHandle = open(outFileName,'w')
        neoObj = pickle.load(fileHandle)
        segments = neoObj.segments
        segment = segments[0]
        spikeTrains = segment.spiketrains
        neurons = len(spikeTrains)
        for neuronNum in range (0,neurons):
            if (len(spikeTrains[neuronNum])>0):
                spikes = spikeTrains[neuronNum]
                for spike in range (0,len(spikes)):
                    outFileHandle.write(str( neuronNum) + " " + str(spikes[spike])+ "\n")
        outFileHandle.close()
        fileHandle.close()

def makeOneTrainSynapses(spikeSource,verbHier,np1Hier,prepCells,np2Hier,
                         attachNCells,attachVCells,tuple):
    verbName = tuple[0]
    np1Name = tuple[1]
    prepName = tuple[2]
    np2Name = tuple[3]
    verbCA = verbHier.getUnitNumber(verbName)
    np1CA = np1Hier.getUnitNumber(np1Name)
    prepCA = getPrepOffset(prepName)
    np2CA = np2Hier.getUnitNumber(np2Name)
    makeSynapsesFromSourceToCA(spikeSource,verbHier.cells,verbCA*neuronsPerCA)
    makeSynapsesFromSourceToCA(spikeSource,np1Hier.cells,np1CA*neuronsPerCA)
    makeSynapsesFromSourceToCA(spikeSource,prepCells,prepCA*neuronsPerCA)
    makeSynapsesFromSourceToCA(spikeSource,np2Hier.cells,np2CA*neuronsPerCA)
    if (tuple[4] == 'np'):
        makeSynapsesFromSourceToCA(spikeSource,attachNCells,0)
    else:
        makeSynapsesFromSourceToCA(spikeSource,attachVCells,0)

def makeStops(numberTests,verbHier,np1Hier,np2Hier):
    stopTimeStart = 160
    stopTimes = getStopTrainSpikeTimes(stopTimeStart,numberTests,
                                       timeBetweenPresentations)
    print stopTimes
    presentArray = {'spike_times': [stopTimes]}
    inhibitSpikeGen=sim.Population(1,sim.SpikeSourceArray,presentArray)
    sourceStopsPopulation(inhibitSpikeGen,verbHier.cells,verbHier.cells.size)
    sourceStopsPopulation(inhibitSpikeGen,np1Hier.cells,np1Hier.cells.size)
    sourceStopsPopulation(inhibitSpikeGen,np2Hier.cells,np2Hier.cells.size)

def makeTrainSynapses(verbHier,np1Hier,prep,np2Hier,attachN,attachV,tuples):
    numberTests = len(tuples)
    makeStops(numberTests,verbHier,np1Hier,np2Hier)
    spikeSources = createCAInputs(numberTests)

    for testNumber in range (0,numberTests):
        print tuples[testNumber]
        makeOneTrainSynapses(spikeSources[testNumber],verbHier,np1Hier,prep,
                             np2Hier,attachN,attachV,tuples[testNumber])

def createHierarchicalNetwork(fileName,nealInst):
    hier = CreateHierClass()
    hier.createNetForCountingNoTraining(fileName,nealInst)
    makeCAs(hier.cells)
    makeHiersFromHier(hier)
    makeCAs(hier.cells)
    return hier

def getPrepOffset(prepName):
    for index in range (0, len(preps)):
        if (preps[index] == prepName):
            return index

    print "getPrepOffset error ", prepName
    return -1

def readPrepFile(prepFileName):
    preps = list()
    prepFileHandle = open(prepFileName,'r')
    line = " "
    done = False
    while (not( len(line) == 0)):
        line = prepFileHandle.readline()
        prep = line.strip()
        if (len(prep) > 0):
            preps.append(prep)

    prepFileHandle.close()
    return preps

def createPrepNetwork(fileName):
    global numberPreps
    global preps
    preps = readPrepFile(fileName)
    print preps
    numberPreps = len(preps)
    prepCells =  sim.Population(10*numberPreps,sim.IF_cond_exp, {})
    prepCells.record(['spikes'])
    return prepCells

def readTestFourTuples(fourTupleFileName):
    result = []
    fileHandle = open(fourTupleFileName,'r')
    done = False
    while (not done):
        line = fileHandle.readline()
        if (line[0] =='@'):
            done = True
            return result
        fiveTuple = line[0:len(line)-1].split(' ')
        result = result + [fiveTuple]

    print "Error, should not be here."
    return -1

#---main--
#args = sys.argv
#numberArgs = args.__len__()
#if (numberArgs == 2):
#    baseFileName = args[1]
#else:

baseFileName = "nov22"

fileName = baseFileName + "Hier"
verbFileName = dirName+"verb"+fileName
np1FileName = dirName+"np1"+fileName 
prepFileName = dirName+"prepCondense"+baseFileName 
np2FileName = dirName+"np2"+fileName
attachNFileName = dirName+"aN"+fileName
attachVFileName = dirName+"aV"+fileName
trainTupleFileName = dirName + baseFileName + "Tuples.txt"

sim.setup(timestep=1.0, max_delay=1.0)
neal = NealCoverFunctions(nealParameters.simulator)

verbHier = createHierarchicalNetwork(verbFileName,neal)
np1Hier = createHierarchicalNetwork(np1FileName,neal)
prepCells = createPrepNetwork(prepFileName+".txt")
np2Hier = createHierarchicalNetwork(np2FileName,neal)

attachToNPCells =  sim.Population(10,sim.IF_cond_exp, {})
attachToNPCells.record(['spikes'])
attachToVPCells =  sim.Population(10,sim.IF_cond_exp, {})
attachToVPCells.record(['spikes'])

tuples = readTestFourTuples(trainTupleFileName)

makeTrainSynapses(verbHier,np1Hier,prepCells,np2Hier,attachToNPCells,
                 attachToVPCells, tuples)

neal.nealApplyProjections()

duration = timeBetweenPresentations*len(tuples)
print duration 
sim.run(duration)

prepSpFile = prepFileName+".sp"
finalVerbSpFile = verbFileName+"Final.sp"
finalNp1SpFile = np1FileName+"Final.sp"
finalNp2SpFile = np2FileName+"Final.sp"
finalAttachNSpFile = attachNFileName+"Final.sp"
finalAttachVSpFile = attachVFileName+"Final.sp"
if (nealParameters.simulator=="spinnaker") and (nealParameters.spinnVersion == 3.0):
    verbHier.cells.printSpikes(finalVerbSpFile)
    np1Hier.cells.printSpikes(finalNp1SpFile)
    np2Hier.cells.printSpikes(finalNp2SpFile)
else:  #nest or spinnaker v4
    prepPklFile = prepFileName+".pkl"
    prepCells.write_data(prepPklFile,'spikes')
    printPklSpikes(prepPklFile,prepSpFile)
    finalVerbPklFile = verbFileName+"Final.pkl"
    verbHier.cells.write_data(finalVerbPklFile,'spikes')
    printPklSpikes(finalVerbPklFile,finalVerbSpFile)
    finalNp1PklFile = np1FileName+"Final.pkl"
    np1Hier.cells.write_data(finalNp1PklFile,'spikes')
    printPklSpikes(finalNp1PklFile,finalNp1SpFile)
    finalNp2PklFile = np2FileName+"Final.pkl"
    np2Hier.cells.write_data(finalNp2PklFile,'spikes')
    printPklSpikes(finalNp2PklFile,finalNp2SpFile)
    
    finalAttachNPklFile = attachNFileName+"Final.pkl"
    attachToNPCells.write_data(finalAttachNPklFile,'spikes')
    printPklSpikes(finalAttachNPklFile,finalAttachNSpFile)
    finalAttachVPklFile = attachVFileName+"Final.pkl"
    attachToVPCells.write_data(finalAttachVPklFile,'spikes')
    printPklSpikes(finalAttachVPklFile,finalAttachVSpFile)
