
import nealParams as nealParameters
if (nealParameters.simulator=="spinnaker"):
    import pyNN.spiNNaker as sim
elif (nealParameters.simulator=="nest"):
    import pyNN.nest as sim

from nealCoverClass import NealCoverFunctions

class CreateHierClass:
    #constants
    numCATrainingEpochs = 16
    sourceToNeuronWt = 0.1
    cellsPerCA = 10

    #instance variables
    units = []
    numberUnits = -1
    isARelationships = []
    numberIsAs = -1
    duration = -1
    cells = "uninitialized nest neurons"

    def init(self,nealInstance):
        self.neal = nealInstance

    def getUnitNumber(self,name):
        nameIndex = unit.index(name)
        return nameIndex

    #read lines until you get one starting with an @
    #each of these is a primitive unit in the hierarchy
    def readUnits(self,handle):
        unitListDone = False
        while (not unitListDone):
            line = handle.readline()
            if (line[0]=='@'):
                unitListDone = True
            else:
                unitName = line.strip() #take off the \n
                self.units = self.units + [unitName]
    
    def inUnits(self,checkUnit):
        done = False
        unitListOffset = 0
        while (not done):
            if (checkUnit == self.units[unitListOffset]):
                return True
            unitListOffset = unitListOffset + 1
            if (unitListOffset == self.numberUnits):
                done = True
        return False

    def getUnitNumber(self,checkUnit):
        for resultUnit in range (0,self.numberUnits):
            if (checkUnit == self.units[resultUnit]):
                return resultUnit

        print "error ", checkUnit , " not in unit array"

    #read lines until you get one starting with an @
    #each line should be three tuple of A isA B
    #report error if it's not a threeTuple, if A or B is not in units and 
    #isA is not isA
    def readIsAs(self,handle):
        relationshipListDone = False
        while (not relationshipListDone):
            line = handle.readline()
            if (line[0]=='@'):
                relationshipListDone = True
            else:
                stringTriplet = line.split(' ')
                if (len(stringTriplet) != 3):
                    print "error ", line, " wrong length"
                if (stringTriplet[1] != "isA"):
                    print "error ", line, " only handling isA now"
                subCatName = stringTriplet[0]
                superCatName = stringTriplet[2]
                superCatName = superCatName.strip()
                #Check sub and supercat in units
                if (not self.inUnits(subCatName)):
                    print "error ", subCatName, " not a unit"
                if (not self.inUnits(superCatName)):
                    print "error ", superCatName, " not a unit"

                self.isARelationships = self.isARelationships + [
                    [subCatName,superCatName]]

    def createViaRead(self,fileName):
        fileHandle = open(fileName, 'r')
        self.readUnits(fileHandle)
        self.numberUnits = len(self.units)
        self.readIsAs(fileHandle)
        self.numberIsAs = len(self.isARelationships)
        fileHandle.close()

    def setup(self):
        sim.setup(timestep=1.0, max_delay=1.0)

    def createNeurons(self):
        numCells = self.numberUnits * self.cellsPerCA
        print "nc", numCells
        self.cells = sim.Population(numCells,sim.IF_cond_exp, {})
        self.cells.record(['spikes'])

    def getTrainSpikeTimes(self,start, step, steps):
        inputSpikeTimes = []
        for numTrainSpikes in range (0,steps): 
            nextTime = start + (numTrainSpikes *step) 
            inputSpikeTimes = inputSpikeTimes + [nextTime]
        return inputSpikeTimes

    def makeSynapsesFromSourceToCA(self,spikeSource,cells,toStart):
        inputWeight = self.sourceToNeuronWt
        connector = []
        for toOffset in range (0,self.cellsPerCA):
            toNeuron = toOffset +toStart
            connector=connector+[(0,toNeuron,inputWeight,nealParameters.DELAY)]
        self.neal.nealProjection(spikeSource, cells, connector,'excitatory')

    #for each unit, make a spikeSource for training
    #then create synapses from that spikeSource to the neurons in the unit.
    def createCATrainingSynapses(self):
        timeBetweenCASpikes = 50
        timeBetweenCAEpochs = (self.numberUnits + 1)*timeBetweenCASpikes

        for unitNumber in range (0,self.numberUnits):
            firstSpike = (unitNumber*timeBetweenCASpikes) + 10
            inputSpikeTimes = self.getTrainSpikeTimes(firstSpike,
                    timeBetweenCAEpochs,self.numCATrainingEpochs)
            spikeArray = {'spike_times': [inputSpikeTimes]}
            spikeSource = sim.Population(1,sim.SpikeSourceArray,spikeArray)

            self.makeSynapsesFromSourceToCA(spikeSource,self.cells,
                                           unitNumber*self.cellsPerCA)

            #print inputSpikeTimes
        CATrainingDuration = (timeBetweenCAEpochs*self.numCATrainingEpochs)

        #print CATrainingDuration 
        return CATrainingDuration

    #For each association pair, create a spikeSource and 
    #synapses from the source to the pair's cells.
    def createAssociationPairSynapses(self,startTime):
        timeBetweenAssociationSpikes = 50
        for pairNumber in range (0,self.numberIsAs):
            firstSpike = startTime+(pairNumber*timeBetweenAssociationSpikes)+10
            #this assumes just one spike per pair
            inputSpikeTimes = self.getTrainSpikeTimes(firstSpike,1,1)
            spikeArray = {'spike_times': [inputSpikeTimes]}
            spikeSource = sim.Population(1,sim.SpikeSourceArray,spikeArray)
            #print inputSpikeTimes, self.isARelationships[pairNumber]

            unitPair = self.isARelationships[pairNumber]
            subCatUnit = unitPair[0]
            superCatUnit = unitPair[1]
            subCatNumber = self.getUnitNumber(subCatUnit)
            superCatNumber = self.getUnitNumber(superCatUnit)
            #print subCatNumber,superCatNumber

            self.makeSynapsesFromSourceToCA(spikeSource,self.cells,
                                           subCatNumber*self.cellsPerCA)
            self.makeSynapsesFromSourceToCA(spikeSource,self.cells,
                                           superCatNumber*self.cellsPerCA)

        AssociationTrainingDuration = (timeBetweenAssociationSpikes*
                                      (self.numberIsAs + 1))

        #print AssociationTrainingDuration
        return AssociationTrainingDuration

    def getUnitTriples(self):
        unitTriples = []
        for unitPair in self.isARelationships:
            parentCatUnit = unitPair[1]
            for parentPair in self.isARelationships:
                potentialParentCatUnit = parentPair[0]
                if (parentCatUnit == potentialParentCatUnit):
                    childCatUnit = unitPair[0]
                    grandParentUnit = parentPair[1]
                    #print childCatUnit,parentCatUnit, grandParentUnit
                    unitTriples = unitTriples + [[childCatUnit,parentCatUnit, 
                                                 grandParentUnit]]

        #print unitTriples
        return unitTriples

    def createAssociationTripleSynapses(self,startTime):
        unitTriples = self.getUnitTriples()
        numTriples = len(unitTriples)
        timeBetweenAssociationSpikes = 50
        for tripleNumber in range (0,numTriples):
            firstSpike =startTime+(tripleNumber*timeBetweenAssociationSpikes)+10
            inputSpikeTimes = self.getTrainSpikeTimes(firstSpike,1,1)
            spikeArray = {'spike_times': [inputSpikeTimes]}
            spikeSource = sim.Population(1,sim.SpikeSourceArray,spikeArray)
            #print inputSpikeTimes, unitTriples[tripleNumber]
            unitTriple = unitTriples[tripleNumber]
            childCatUnit = unitTriple[0]
            parentCatUnit = unitTriple[1]
            grandParentCatUnit = unitTriple[2]
            childCatNumber = self.getUnitNumber(childCatUnit)
            parentCatNumber = self.getUnitNumber(parentCatUnit)
            grandParentCatNumber = self.getUnitNumber(grandParentCatUnit)
            #print childCatNumber,parentCatNumber,grandParentCatNumber
            self.makeSynapsesFromSourceToCA(spikeSource,self.cells,
                                           childCatNumber*self.cellsPerCA)
            self.makeSynapsesFromSourceToCA(spikeSource,self.cells,
                                           parentCatNumber*self.cellsPerCA)
            self.makeSynapsesFromSourceToCA(spikeSource,self.cells,
                                           grandParentCatNumber*self.cellsPerCA)


        tripleTrainingDuration = (timeBetweenAssociationSpikes*(numTriples + 1))
        return tripleTrainingDuration


    def createAssociationTrainingSynapses(self,startTime):
        pairDuration = self.createAssociationPairSynapses(startTime)
        tripleStartTime = startTime + pairDuration
        tripleDuration = self.createAssociationTripleSynapses(tripleStartTime)
        return pairDuration+tripleDuration

    def createTrainingSynapses(self):
        CATrainingDuration = self.createCATrainingSynapses()
        associationTrainingDuration = self.createAssociationTrainingSynapses(
            CATrainingDuration)

        self.duration = CATrainingDuration + associationTrainingDuration

    def runTraining(self):
        #print self.duration
        sim.run(self.duration)

    def storeResults(self,fileName):
        storeFileName = 'results/'+fileName+'.pkl'
        self.cells.write_data(storeFileName,'spikes')

    def createNetForCounting(self,fileName,nealInst):
        self.init(nealInst)
        inputFileName = fileName+".txt"
        self.createViaRead(inputFileName)
        self.setup()
        self.createNeurons()
        self.createTrainingSynapses()
        self.runTraining()
        self.storeResults(fileName)

    def createNetForCountingNoTraining(self,fileName,nealInst):
        self.init(nealInst)
        inputFileName = fileName+".txt"
        self.createViaRead(inputFileName)
        self.createNeurons()

